@@ -120,11 +120,19 @@ impl PhysicalExpr for AggregationExpr {
             } else {
                 s.head(Some(1))
             }),
+            GroupByMethod::FirstNonNull => Ok(s
+                .as_materialized_series_maintain_scalar()
+                .first_non_null()
+                .into_column(s.name().clone())),
             GroupByMethod::Last => Ok(if s.is_empty() {
                 Column::full_null(s.name().clone(), 1, s.dtype())
             } else {
                 s.tail(Some(1))
             }),
+            GroupByMethod::LastNonNull => Ok(s
+                .as_materialized_series_maintain_scalar()
+                .last_non_null()
+                .into_column(s.name().clone())),
             GroupByMethod::Item { allow_empty } => Ok(match s.len() {
                 0 if allow_empty => Column::full_null(s.name().clone(), 1, s.dtype()),
                 1 => s,
@@ -332,11 +340,21 @@ impl PhysicalExpr for AggregationExpr {
                     let agg_s = s.agg_first(&groups);
                     AggregatedScalar(agg_s.with_name(keep_name))
                 },
+                GroupByMethod::FirstNonNull => {
+                    let (s, groups) = ac.get_final_aggregation();
+                    let agg_s = s.agg_first_non_null(&groups);
+                    AggregatedScalar(agg_s.with_name(keep_name))
+                },
                 GroupByMethod::Last => {
                     let (s, groups) = ac.get_final_aggregation();
                     let agg_s = s.agg_last(&groups);
                     AggregatedScalar(agg_s.with_name(keep_name))
                 },
+                GroupByMethod::LastNonNull => {
+                    let (s, groups) = ac.get_final_aggregation();
+                    let agg_s = s.agg_last_non_null(&groups);
+                    AggregatedScalar(agg_s.with_name(keep_name))
+                },
                 GroupByMethod::Item { allow_empty } => {
                     let (s, groups) = ac.get_final_aggregation();
                     for gc in groups.group_count().iter() {
