@@ -2,8 +2,9 @@ from __future__ import annotations
 
 import typing
 from collections import OrderedDict
-from datetime import date, datetime, timedelta
+from datetime import date, datetime, time, timedelta
 from typing import TYPE_CHECKING, Any
+from zoneinfo import ZoneInfo
 
 import numpy as np
 import pytest
@@ -12,15 +13,18 @@ from hypothesis import given
 import polars as pl
 import polars.selectors as cs
 from polars import Expr
-from polars.exceptions import ColumnNotFoundError
+from polars.exceptions import (
+    ColumnNotFoundError,
+    InvalidOperationError,
+)
 from polars.meta import get_index_type
 from polars.testing import assert_frame_equal, assert_series_equal
-from polars.testing.parametric import series
+from polars.testing.parametric import column, dataframes, series
 
 if TYPE_CHECKING:
     from typing import Callable
 
-    from polars._typing import PolarsDataType
+    from polars._typing import PolarsDataType, TimeUnit
 
 
 def test_group_by() -> None:
@@ -304,6 +308,190 @@ def test_group_by_shorthand_quantile(df: pl.DataFrame) -> None:
     assert result.rows() == expected
 
 
+def test_group_by_quantile_date() -> None:
+    df = pl.DataFrame(
+        {
+            "group": [1, 1, 1, 1, 2, 2, 2, 2],
+            "value": [date(2025, 1, x) for x in range(1, 9)],
+        }
+    )
+    result = (
+        df.lazy()
+        .group_by("group", maintain_order=True)
+        .agg(
+            nearest=pl.col("value").quantile(0.5, "nearest"),
+            higher=pl.col("value").quantile(0.5, "higher"),
+            lower=pl.col("value").quantile(0.5, "lower"),
+            linear=pl.col("value").quantile(0.5, "linear"),
+        )
+    )
+    dt = pl.Datetime("us")
+    expected = pl.DataFrame(
+        {
+            "group": [1, 2],
+            "nearest": pl.Series(
+                [datetime(2025, 1, 3), datetime(2025, 1, 7)], dtype=dt
+            ),
+            "higher": pl.Series([datetime(2025, 1, 3), datetime(2025, 1, 7)], dtype=dt),
+            "lower": pl.Series([datetime(2025, 1, 2), datetime(2025, 1, 6)], dtype=dt),
+            "linear": pl.Series(
+                [datetime(2025, 1, 2, 12), datetime(2025, 1, 6, 12)], dtype=dt
+            ),
+        }
+    )
+    assert result.collect_schema() == pl.Schema(
+        {  # type: ignore[arg-type]
+            "group": pl.Int64,
+            "nearest": dt,
+            "higher": dt,
+            "lower": dt,
+            "linear": dt,
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+
+@pytest.mark.parametrize("tu", ["ms", "us", "ns"])
+@pytest.mark.parametrize("time_zone", [None, "Asia/Tokyo"])
+def test_group_by_quantile_datetime(tu: TimeUnit, time_zone: str) -> None:
+    dt = pl.Datetime(tu, time_zone)
+    tz = ZoneInfo(time_zone) if time_zone else None
+    df = pl.DataFrame(
+        {
+            "group": [1, 1, 1, 1, 2, 2, 2, 2],
+            "value": pl.Series(
+                [datetime(2025, 1, x, tzinfo=tz) for x in range(1, 9)],
+                dtype=dt,
+            ),
+        }
+    )
+    result = (
+        df.lazy()
+        .group_by("group", maintain_order=True)
+        .agg(
+            nearest=pl.col("value").quantile(0.5, "nearest"),
+            higher=pl.col("value").quantile(0.5, "higher"),
+            lower=pl.col("value").quantile(0.5, "lower"),
+            linear=pl.col("value").quantile(0.5, "linear"),
+        )
+    )
+    expected = pl.DataFrame(
+        {
+            "group": [1, 2],
+            "nearest": pl.Series(
+                [datetime(2025, 1, 3, tzinfo=tz), datetime(2025, 1, 7, tzinfo=tz)],
+                dtype=dt,
+            ),
+            "higher": pl.Series(
+                [datetime(2025, 1, 3, tzinfo=tz), datetime(2025, 1, 7, tzinfo=tz)],
+                dtype=dt,
+            ),
+            "lower": pl.Series(
+                [datetime(2025, 1, 2, tzinfo=tz), datetime(2025, 1, 6, tzinfo=tz)],
+                dtype=dt,
+            ),
+            "linear": pl.Series(
+                [
+                    datetime(2025, 1, 2, 12, tzinfo=tz),
+                    datetime(2025, 1, 6, 12, tzinfo=tz),
+                ],
+                dtype=dt,
+            ),
+        }
+    )
+    assert result.collect_schema() == pl.Schema(
+        {  # type: ignore[arg-type]
+            "group": pl.Int64,
+            "nearest": dt,
+            "higher": dt,
+            "lower": dt,
+            "linear": dt,
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+
+@pytest.mark.parametrize("tu", ["ms", "us", "ns"])
+def test_group_by_quantile_duration(tu: TimeUnit) -> None:
+    dt = pl.Duration(tu)
+    df = pl.DataFrame(
+        {
+            "group": [1, 1, 1, 1, 2, 2, 2, 2],
+            "value": pl.Series([timedelta(hours=x) for x in range(1, 9)], dtype=dt),
+        }
+    )
+    result = (
+        df.lazy()
+        .group_by("group", maintain_order=True)
+        .agg(
+            nearest=pl.col("value").quantile(0.5, "nearest"),
+            higher=pl.col("value").quantile(0.5, "higher"),
+            lower=pl.col("value").quantile(0.5, "lower"),
+            linear=pl.col("value").quantile(0.5, "linear"),
+        )
+    )
+    expected = pl.DataFrame(
+        {
+            "group": [1, 2],
+            "nearest": pl.Series([timedelta(hours=3), timedelta(hours=7)], dtype=dt),
+            "higher": pl.Series([timedelta(hours=3), timedelta(hours=7)], dtype=dt),
+            "lower": pl.Series([timedelta(hours=2), timedelta(hours=6)], dtype=dt),
+            "linear": pl.Series(
+                [timedelta(hours=2, minutes=30), timedelta(hours=6, minutes=30)],
+                dtype=dt,
+            ),
+        }
+    )
+    assert result.collect_schema() == pl.Schema(
+        {  # type: ignore[arg-type]
+            "group": pl.Int64,
+            "nearest": dt,
+            "higher": dt,
+            "lower": dt,
+            "linear": dt,
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+
+def test_group_by_quantile_time() -> None:
+    df = pl.DataFrame(
+        {
+            "group": [1, 1, 1, 1, 2, 2, 2, 2],
+            "value": pl.Series([time(hour=x) for x in range(1, 9)]),
+        }
+    )
+    result = (
+        df.lazy()
+        .group_by("group", maintain_order=True)
+        .agg(
+            nearest=pl.col("value").quantile(0.5, "nearest"),
+            higher=pl.col("value").quantile(0.5, "higher"),
+            lower=pl.col("value").quantile(0.5, "lower"),
+            linear=pl.col("value").quantile(0.5, "linear"),
+        )
+    )
+    expected = pl.DataFrame(
+        {
+            "group": [1, 2],
+            "nearest": pl.Series([time(hour=3), time(hour=7)]),
+            "higher": pl.Series([time(hour=3), time(hour=7)]),
+            "lower": pl.Series([time(hour=2), time(hour=6)]),
+            "linear": pl.Series([time(hour=2, minute=30), time(hour=6, minute=30)]),
+        }
+    )
+    assert result.collect_schema() == pl.Schema(
+        {
+            "group": pl.Int64,
+            "nearest": pl.Time,
+            "higher": pl.Time,
+            "lower": pl.Time,
+            "linear": pl.Time,
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+
 def test_group_by_args() -> None:
     df = pl.DataFrame(
         {
@@ -415,7 +603,7 @@ def test_group_by_sorted_empty_dataframe_3680() -> None:
     )
     assert df.rows() == []
     assert df.shape == (0, 2)
-    assert df.schema == {"key": pl.Categorical(ordering="lexical"), "val": pl.Float64}
+    assert df.schema == {"key": pl.Categorical(), "val": pl.Float64}
 
 
 def test_group_by_custom_agg_empty_list() -> None:
@@ -1474,6 +1662,24 @@ def test_group_by_shift_filter_23910(maintain_order: bool) -> None:
     )
 
 
+@pytest.mark.parametrize("maintain_order", [False, True])
+def test_group_by_having(maintain_order: bool) -> None:
+    df = pl.DataFrame(
+        {
+            "grp": ["A", "A", "B", "B", "C", "C"],
+            "value": [10, 15, 5, 15, 5, 10],
+        }
+    )
+
+    result = (
+        df.group_by("grp", maintain_order=maintain_order)
+        .having(pl.col("value").mean() >= 10)
+        .agg()
+    )
+    expected = pl.DataFrame({"grp": ["A", "B"]})
+    assert_frame_equal(result, expected, check_row_order=maintain_order)
+
+
 def test_group_by_tuple_typing_24112() -> None:
     df = pl.DataFrame({"id": ["a", "b", "a"], "val": [1, 2, 3]})
     for (id_,), _ in df.group_by("id"):
@@ -1886,6 +2092,7 @@ def test_group_by_any_all(expr: Callable[[pl.Expr], pl.Expr]) -> None:
         allow_chunks=False,  # bug: See #24960
     )
 )
+@pytest.mark.may_fail_auto_streaming  # bug: See #24960
 def test_group_by_skew_kurtosis(s: pl.Series) -> None:
     df = s.to_frame()
 
@@ -2130,3 +2337,586 @@ def test_group_by_drop_nans(s: pl.Series) -> None:
             pl.col.f.implode().list.eval(pl.element().drop_nans()).reshape((-1,))
         ),
     )
+
+
+@given(
+    df=dataframes(
+        min_size=1,
+        include_cols=[column(name="key", dtype=pl.UInt8, allow_null=False)],
+    ),
+)
+@pytest.mark.parametrize(
+    ("expr", "check_order", "returns_scalar", "length_preserving", "is_window"),
+    [
+        (pl.Expr.unique, False, False, False, False),
+        (lambda e: e.unique(maintain_order=True), True, False, False, False),
+        (pl.Expr.drop_nans, True, False, False, False),
+        (pl.Expr.drop_nulls, True, False, False, False),
+        (pl.Expr.null_count, True, False, False, False),
+        (pl.Expr.n_unique, True, True, False, False),
+        (
+            lambda e: e.filter(pl.int_range(0, e.len()) % 3 == 0),
+            True,
+            False,
+            False,
+            False,
+        ),
+        (pl.Expr.shift, True, False, True, False),
+        (pl.Expr.forward_fill, True, False, True, False),
+        (pl.Expr.backward_fill, True, False, True, False),
+        (pl.Expr.reverse, True, False, True, False),
+        (
+            lambda e: (pl.int_range(e.len() - e.len(), e.len()) % 3 == 0).any(),
+            True,
+            True,
+            False,
+            False,
+        ),
+        (
+            lambda e: (pl.int_range(e.len() - e.len(), e.len()) % 3 == 0).all(),
+            True,
+            True,
+            False,
+            False,
+        ),
+        (lambda e: e.head(2), True, False, False, False),
+        (pl.Expr.first, True, True, False, False),
+        (pl.Expr.mode, False, False, False, False),
+        (lambda e: e.fill_null(e.first()).over(e), True, False, True, True),
+        (lambda e: e.first().over(e), True, False, True, True),
+        (
+            lambda e: e.fill_null(e.first()).over(e, mapping_strategy="join"),
+            True,
+            False,
+            True,
+            True,
+        ),
+        (
+            lambda e: e.fill_null(e.first()).over(e, mapping_strategy="explode"),
+            True,
+            False,
+            False,
+            True,
+        ),
+        (
+            lambda e: e.fill_null(strategy="forward").over([e, e]),
+            True,
+            False,
+            True,
+            True,
+        ),
+        (lambda e: e.fill_null(e.first()).over(e, order_by=e), True, False, True, True),
+        (
+            lambda e: e.fill_null(e.first()).over(e, order_by=e, descending=True),
+            True,
+            False,
+            True,
+            True,
+        ),
+        (
+            lambda e: e.gather(pl.int_range(0, e.len()).slice(1, 3)),
+            True,
+            False,
+            False,
+            False,
+        ),
+    ],
+)
+def test_grouped_agg_parametric(
+    df: pl.DataFrame,
+    expr: Callable[[pl.Expr], pl.Expr],
+    check_order: bool,
+    returns_scalar: bool,
+    length_preserving: bool,
+    is_window: bool,
+) -> None:
+    types: dict[str, tuple[Callable[[pl.Expr], pl.Expr], bool, bool]] = {
+        "basic": (lambda e: e, False, True),
+    }
+
+    if not is_window:
+        types["first"] = (pl.Expr.first, True, False)
+        types["slice"] = (lambda e: e.slice(1, 3), False, False)
+        types["impl_expl"] = (lambda e: e.implode().explode(), False, False)
+        types["rolling"] = (
+            lambda e: e.rolling(pl.row_index(), period="3i"),
+            False,
+            True,
+        )
+        types["over"] = (lambda e: e.forward_fill().over(e), False, True)
+
+    def slit(s: pl.Series) -> pl.Expr:
+        import polars._plr as plr
+
+        return pl.Expr._from_pyexpr(plr.lit(s._s, False, is_scalar=True))
+
+    df = df.with_columns(pl.col.key % 4)
+    gb = df.group_by("key").agg(
+        *[
+            expr(t(~cs.by_name("key"))).name.prefix(f"{k}_")
+            for k, (t, _, _) in types.items()
+        ],
+        *[
+            expr(slit(df[c].head(1))).alias(f"literal_{c}")
+            for c in filter(lambda c: c != "key", df.columns)
+        ],
+    )
+    ls = (
+        df.group_by("key")
+        .agg(pl.all())
+        .select(
+            pl.col.key,
+            *[
+                (~cs.by_name("key"))
+                .list.agg(expr(t(pl.element())))
+                .name.prefix(f"{k}_")
+                for k, (t, _, _) in types.items()
+            ],
+            *[
+                pl.col(c).list.agg(expr(slit(df[c].head(1)))).alias(f"literal_{c}")
+                for c in filter(lambda c: c != "key", df.columns)
+            ],
+        )
+    )
+
+    if not is_window:
+        types["literal"] = (lambda e: e, True, False)
+
+    def verify_index(i: int) -> None:
+        idx_df = df.filter(pl.col.key == pl.lit(i, pl.UInt8))
+        idx_gb = gb.filter(pl.col.key == pl.lit(i, pl.UInt8))
+        idx_ls = ls.filter(pl.col.key == pl.lit(i, pl.UInt8))
+
+        for col in df.columns:
+            if col == "key":
+                continue
+
+            for k, (t, t_is_scalar, t_is_length_preserving) in types.items():
+                c = f"{k}_{col}"
+
+                if k == "literal":
+                    df_s = idx_df.select(
+                        expr(t(slit(df[col].head(1)))).alias(c)
+                    ).to_series()
+                else:
+                    df_s = idx_df.select(expr(t(pl.col(col))).alias(c)).to_series()
+
+                gb_s = idx_gb[c]
+                ls_s = idx_ls[c]
+
+                result_is_scalar = False
+                result_is_scalar |= returns_scalar and t_is_length_preserving
+                result_is_scalar |= t_is_scalar and length_preserving
+                result_is_scalar &= not is_window
+
+                if not result_is_scalar:
+                    gb_s = gb_s.explode(empty_as_null=False)
+                    ls_s = ls_s.explode(empty_as_null=False)
+
+                assert_series_equal(df_s, gb_s, check_order=check_order)
+                assert_series_equal(df_s, ls_s, check_order=check_order)
+
+    if 0 in df["key"]:
+        verify_index(0)
+    if 1 in df["key"]:
+        verify_index(1)
+    if 2 in df["key"]:
+        verify_index(2)
+    if 3 in df["key"]:
+        verify_index(3)
+
+
+@pytest.mark.parametrize("maintain_order", [False, True])
+@pytest.mark.parametrize(
+    ("df", "out"),
+    [
+        (
+            pl.DataFrame(
+                {
+                    "key": [0, 0, 0, 0, 1],
+                    "a": [True, False, False, False, False],
+                }
+            ).with_columns(
+                a=pl.when(pl.Series([False, False, False, False, True])).then(pl.col.a)
+            ),
+            pl.DataFrame(
+                {
+                    "key": [0, 1],
+                    "a": [1, 1],
+                },
+                schema_overrides={"a": pl.get_index_type()},
+            ),
+        ),
+        (
+            pl.DataFrame(
+                {
+                    "key": [0, 0, 1, 1],
+                    "a": [False, False, False, False],
+                }
+            ).with_columns(
+                a=pl.when(pl.Series([False, False, True, True])).then(pl.col.a)
+            ),
+            pl.DataFrame(
+                {
+                    "key": [0, 1],
+                    "a": [1, 1],
+                },
+                schema_overrides={"a": pl.get_index_type()},
+            ),
+        ),
+    ],
+)
+def test_n_unique_masked_bools(
+    maintain_order: bool, df: pl.DataFrame, out: pl.DataFrame
+) -> None:
+    df = df
+
+    assert_frame_equal(
+        df.group_by("key", maintain_order=maintain_order).agg(pl.col.a.n_unique()),
+        out,
+        check_row_order=maintain_order,
+    )
+    assert_frame_equal(
+        df.group_by("key", maintain_order=maintain_order)
+        .agg(pl.col.a)
+        .with_columns(pl.col.a.list.agg(pl.element().n_unique())),
+        out,
+        check_row_order=maintain_order,
+    )
+
+
+@pytest.mark.parametrize("maintain_order", [False, True])
+@pytest.mark.parametrize("stable", [False, True])
+def test_group_bool_unique_25267(maintain_order: bool, stable: bool) -> None:
+    df = pl.DataFrame(
+        {
+            "id": ["A", "A", "B", "B", "C", "C"],
+            "str_values": ["D", "E", "F", "F", "G", "G"],
+            "bool_values": [True, False, True, True, False, False],
+        }
+    )
+
+    gb = df.group_by("id", maintain_order=maintain_order).agg(
+        pl.col("str_values", "bool_values").unique(maintain_order=stable),
+    )
+
+    ls = (
+        df.group_by("id", maintain_order=maintain_order)
+        .agg("str_values", "bool_values")
+        .with_columns(
+            pl.col("str_values", "bool_values").list.agg(
+                pl.element().unique(maintain_order=stable)
+            )
+        )
+    )
+
+    for i in ["A", "B", "C"]:
+        for c in ["str_values", "bool_values"]:
+            df_s = (
+                df.select(pl.col(c).filter(pl.col.id == pl.lit(i)))
+                .to_series()
+                .unique(maintain_order=stable)
+            )
+            gb_s = gb.select(
+                pl.col(c).filter(pl.col.id == pl.lit(i)).reshape((-1,))
+            ).to_series()
+            ls_s = ls.select(
+                pl.col(c).filter(pl.col.id == pl.lit(i)).reshape((-1,))
+            ).to_series()
+
+            assert_series_equal(df_s, gb_s, check_order=stable)
+            assert_series_equal(df_s, ls_s, check_order=stable)
+
+
+@pytest.mark.parametrize("group_as_slice", [False, True])
+@pytest.mark.parametrize("n", [10, 100, 1_000, 10_000])
+@pytest.mark.parametrize(
+    "dtype", [pl.Int32, pl.Boolean, pl.String, pl.Categorical, pl.List(pl.Int32)]
+)
+def test_group_by_first_last(
+    group_as_slice: bool, n: int, dtype: PolarsDataType
+) -> None:
+    idx = pl.Series([1, 2, 3, 4, 5], dtype=pl.Int32)
+
+    lf = pl.LazyFrame(
+        {
+            "idx": pl.Series(
+                [1] * n + [2] * n + [3] * n + [4] * n + [5] * n, dtype=pl.Int32
+            ),
+            # Each successive group has an additional None spanning the elements
+            "a": pl.Series(
+                [
+                    *[None] * 0, *list(range(1, n + 1)), *[None] * 0,  # idx = 1
+                    *[None] * 1, *list(range(2, n - 0)), *[None] * 1,  # idx = 2
+                    *[None] * 2, *list(range(3, n - 1)), *[None] * 2,  # idx = 3
+                    *[None] * 3, *list(range(4, n - 2)), *[None] * 3,  # idx = 4
+                    *[None] * 4, *list(range(5, n - 3)), *[None] * 4,  # idx = 5
+                ],
+                dtype=pl.Int32,
+            ),
+        }
+    )  # fmt: skip
+    if group_as_slice:
+        lf = lf.set_sorted("idx")  # Use GroupSlice path
+
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        lf = lf.with_columns(pl.col("a").cast(pl.String))
+    lf = lf.with_columns(pl.col("a").cast(dtype))
+
+    # first()
+    result = lf.group_by("idx", maintain_order=True).agg(pl.col("a").first()).collect()
+    expected_vals = pl.Series([1, None, None, None, None])
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        expected_vals = expected_vals.cast(pl.String)
+
+    expected_vals = expected_vals.cast(dtype)
+    expected = pl.DataFrame({"idx": idx, "a": expected_vals})
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).first().collect()
+    assert_frame_equal(result, expected)
+
+    # first(ignore_nulls=True)
+    result = (
+        lf.group_by("idx", maintain_order=True)
+        .agg(pl.col("a").first(ignore_nulls=True))
+        .collect()
+    )
+    expected_vals = pl.Series([1, 2, 3, 4, 5])
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        expected_vals = expected_vals.cast(pl.String)
+
+    expected_vals = expected_vals.cast(dtype)
+    expected = pl.DataFrame({"idx": idx, "a": expected_vals})
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).first(ignore_nulls=True).collect()
+    assert_frame_equal(result, expected)
+
+    # last()
+    result = lf.group_by("idx", maintain_order=True).agg(pl.col("a").last()).collect()
+    expected_vals = pl.Series([n, None, None, None, None])
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        expected_vals = expected_vals.cast(pl.String)
+
+    expected_vals = expected_vals.cast(dtype)
+    expected = pl.DataFrame({"idx": idx, "a": expected_vals})
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).last().collect()
+    assert_frame_equal(result, expected)
+
+    # last_non_null
+    result = (
+        lf.group_by("idx", maintain_order=True)
+        .agg(pl.col("a").last(ignore_nulls=True))
+        .collect()
+    )
+    expected_vals = pl.Series([n, n - 1, n - 2, n - 3, n - 4])
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        expected_vals = expected_vals.cast(pl.String)
+
+    expected_vals = expected_vals.cast(dtype)
+    expected = pl.DataFrame({"idx": idx, "a": expected_vals})
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).last(ignore_nulls=True).collect()
+    assert_frame_equal(result, expected)
+
+    # Test with no nulls
+    lf = pl.LazyFrame(
+        {
+            "idx": pl.Series(
+                [1] * n + [2] * n + [3] * n + [4] * n + [5] * n, dtype=pl.Int32
+            ),
+            # Each successive group has an additional None spanning the elements
+            "a": pl.Series(
+                [
+                    *list(range(1, n + 1)),  # idx = 1
+                    *list(range(2, n + 2)),  # idx = 2
+                    *list(range(3, n + 3)),  # idx = 3
+                    *list(range(4, n + 4)),  # idx = 4
+                    *list(range(5, n + 5)),  # idx = 5
+                ],
+                dtype=pl.Int32,
+            ),
+        }
+    )
+    if group_as_slice:
+        lf = lf.set_sorted("idx")  # Use GroupSlice path
+
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        lf = lf.with_columns(pl.col("a").cast(pl.String))
+    lf = lf.with_columns(pl.col("a").cast(dtype))
+
+    # first()
+    expected_vals = pl.Series([1, 2, 3, 4, 5])
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        expected_vals = expected_vals.cast(pl.String)
+
+    expected_vals = expected_vals.cast(dtype)
+    expected = pl.DataFrame({"idx": idx, "a": expected_vals})
+    result = lf.group_by("idx", maintain_order=True).agg(pl.col("a").first()).collect()
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).first().collect()
+    assert_frame_equal(result, expected)
+
+    # first_non_null
+    result = (
+        lf.group_by("idx", maintain_order=True)
+        .agg(pl.col("a").first(ignore_nulls=True))
+        .collect()
+    )
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).first(ignore_nulls=True).collect()
+    assert_frame_equal(result, expected)
+
+    # last()
+    expected_vals = pl.Series([n, n + 1, n + 2, n + 3, n + 4])
+    if dtype == pl.Categorical:
+        # for Categorical, we must first go through String
+        expected_vals = expected_vals.cast(pl.String)
+
+    expected_vals = expected_vals.cast(dtype)
+    expected = pl.DataFrame({"idx": idx, "a": expected_vals})
+    result = lf.group_by("idx", maintain_order=True).agg(pl.col("a").last()).collect()
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).last().collect()
+    assert_frame_equal(result, expected)
+
+    # last_non_null
+    result = (
+        lf.group_by("idx", maintain_order=True)
+        .agg(pl.col("a").last(ignore_nulls=True))
+        .collect()
+    )
+    assert_frame_equal(result, expected)
+    result = lf.group_by("idx", maintain_order=True).last(ignore_nulls=True).collect()
+    assert_frame_equal(result, expected)
+
+
+def test_sorted_group_by() -> None:
+    lf = pl.LazyFrame(
+        {
+            "a": [1, 1, 2, 2, 3, 3, 3],
+            "b": [4, 5, 8, 1, 0, 1, 3],
+        }
+    )
+
+    lf1 = lf
+    lf2 = lf.set_sorted("a")
+
+    assert_frame_equal(
+        *[
+            q.group_by("a")
+            .agg(b_first=pl.col.b.first(), b_sum=pl.col.b.sum(), b=pl.col.b)
+            .collect(engine="streaming")
+            for q in (lf1, lf2)
+        ],
+        check_row_order=False,
+    )
+
+    lf = lf.with_columns(c=pl.col.a.rle_id())
+    lf1 = lf
+    lf2 = lf.set_sorted("a", "c")
+
+    assert_frame_equal(
+        *[
+            q.group_by("a", "c")
+            .agg(b_first=pl.col.b.first(), b_sum=pl.col.b.sum(), b=pl.col.b)
+            .collect(engine="streaming")
+            for q in (lf1, lf2)
+        ],
+        check_row_order=False,
+    )
+
+
+def test_sorted_group_by_slice() -> None:
+    lf = (
+        pl.DataFrame({"a": [0, 5, 2, 1, 3] * 50})
+        .with_row_index()
+        .with_columns(pl.col.index // 5)
+        .lazy()
+        .set_sorted("index")
+        .group_by("index", maintain_order=True)
+        .agg(pl.col.a.sum() + pl.col.index.first())
+    )
+
+    expected = pl.DataFrame(
+        [
+            pl.Series("index", range(50), pl.get_index_type()),
+            pl.Series("a", range(11, 11 + 50), pl.Int64),
+        ]
+    )
+
+    assert_frame_equal(lf.head(2).collect(), expected.head(2))
+    assert_frame_equal(lf.slice(1, 3).collect(), expected.slice(1, 3))
+    assert_frame_equal(lf.tail(2).collect(), expected.tail(2))
+    assert_frame_equal(lf.slice(5, 1).collect(), expected.slice(5, 1))
+    assert_frame_equal(lf.slice(5, 0).collect(), expected.slice(5, 0))
+    assert_frame_equal(lf.slice(2, 1).collect(), expected.slice(2, 1))
+    assert_frame_equal(lf.slice(50, 1).collect(), expected.slice(50, 1))
+    assert_frame_equal(lf.slice(20, 30).collect(), expected.slice(20, 30))
+    assert_frame_equal(lf.slice(20, 30).collect(), expected.slice(20, 30))
+
+
+def test_agg_first_last_non_null_25405() -> None:
+    lf = pl.LazyFrame(
+        {
+            "a": [1, 1, 1, 1, 2, 2, 2, 2, 2],
+            "b": pl.Series([1, 2, 3, None, None, 4, 5, 6, None]),
+        }
+    )
+
+    # first
+    result = lf.group_by("a", maintain_order=True).agg(
+        pl.col("b").first(ignore_nulls=True)
+    )
+    expected = pl.DataFrame(
+        {
+            "a": [1, 2],
+            "b": [1, 4],
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+    result = lf.with_columns(pl.col("b").first(ignore_nulls=True).over("a"))
+    expected = pl.DataFrame(
+        {
+            "a": [1, 1, 1, 1, 2, 2, 2, 2, 2],
+            "b": [1, 1, 1, 1, 4, 4, 4, 4, 4],
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+    # last
+    result = lf.group_by("a", maintain_order=True).agg(
+        pl.col("b").last(ignore_nulls=True)
+    )
+    expected = pl.DataFrame(
+        {
+            "a": [1, 2],
+            "b": [3, 6],
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+    result = lf.with_columns(pl.col("b").last(ignore_nulls=True).over("a"))
+    expected = pl.DataFrame(
+        {
+            "a": [1, 1, 1, 1, 2, 2, 2, 2, 2],
+            "b": [3, 3, 3, 3, 6, 6, 6, 6, 6],
+        }
+    )
+    assert_frame_equal(result.collect(), expected)
+
+
+def test_group_by_sum_on_strings_should_error_24659() -> None:
+    with pytest.raises(
+        InvalidOperationError,
+        match=r"`sum`.*operation not supported for dtype.*str",
+    ):
+        pl.DataFrame({"str": ["a", "b"]}).group_by(1).agg(pl.col.str.sum())
